# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2022 Jakob Knollmüller, Philipp Arras

import nifty8 as ift
import numpy as np
import resolve as rve
from matplotlib import pyplot as plt

import util as u

try:
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
    master = comm.Get_rank() == 0
except ImportError:
    comm = None
    master = True

# Stage 1: w/o normalization, random start, visibilities
# Stage 2: w/o normalization, random start, closures
# Stage 3: w.  normalization, random start, closures
# Stage 4: w.  normalization, sensible start, closures

obs = u.load_observations("data/SR1_M87_2017_100_hi_hops_netcal_StokesI.uvfits")

space_fov = rve.str2rad("300muas")
space_npix = 256
dx = space_fov / space_npix
sdom = ift.RGSpace([space_npix, space_npix], [dx, dx])

position_space = rve.default_sky_domain(sdom=sdom)

R = rve.InterferometryResponse(obs, position_space, do_wgridding=False, epsilon=1e-6)

CP = u.Visibilities2NormClosurePhases(obs).ducktape(R.target)
CA = u.Visibilities2NormClosureAmplitudes(obs).ducktape(R.target)

correlated_field = ift.SimpleCorrelatedField(
    sdom,
    offset_mean=0,
    offset_std=(1e-3, 1e-6),
    fluctuations=(1.0, 0.8),
    loglogavgslope=(-4.0, 0.5),
    flexibility=None,
    asperity=None,
)
# See webpage for meaning of prior parameters
# https://ift.pages.mpcdf.de/nifty/nifty8/user/getting_started_4_CorrelatedFields.html

sky = correlated_field.exp().ducktape_left(R.domain)
normalization = u.normalization_operator(sky.target)
normalized_sky = normalization @ sky

CP_likelihood = ift.GaussianEnergy(data=CP(obs.vis)) @ CP
CA_likelihood = ift.GaussianEnergy(data=CA(obs.vis)) @ CA

signal = normalized_sky.ducktape_left(R.domain)
likelihood = (CA_likelihood + CP_likelihood) @ R @ signal

# Minimization parameters
ic_sampling = ift.AbsDeltaEnergyController(
    name="Sampling (linear)", deltaE=0.05, iteration_limit=100
)
ic_newton = ift.AbsDeltaEnergyController(
    name="Newton", deltaE=0.5, convergence_level=2, iteration_limit=10
)
ic_lbfgs = ift.AbsDeltaEnergyController(
    name="lbfgs", deltaE=0.5, convergence_level=2, iteration_limit=20
)
minimizer_newton = ift.NewtonCG(ic_newton)
minimizer_lbfgs = ift.VL_BFGS(ic_lbfgs)

# Initial position
initial_position = u.get_initial_position(normalized_sky, ic_sampling, comm)

# Minimize KL
n_iterations = 20
n_samples = lambda iiter: 10 if iiter < 10 else 20
minimizer = lambda iiter: minimizer_lbfgs if iiter < 10 else minimizer_newton


def inspect_callback(sl, iglobal):
    plt.imsave(
        f"sky_{iglobal}.png", sl.average(sky).val.T, origin="lower", cmap="afmhot"
    )


samples = ift.optimize_kl(
    likelihood,
    n_iterations,
    n_samples,
    minimizer,
    ic_sampling,
    None,
    initial_position=initial_position,
    export_operator_outputs={"sky": sky},
    output_directory="results",
    inspect_callback=inspect_callback,
    comm=comm,
)
